"""
Tests for fetching org metadata.

"""

from unittest import TestCase, mock

from runeq.config import Config
from runeq.resources.client import GraphClient
from runeq.resources.event import (
    Event,
    EventSet,
    _reformat_event,
    get_patient_activity_events,
    get_patient_events,
    get_patient_medication_events,
    get_patient_symptom_events,
    get_patient_wellbeing_events,
)


class TestEvent(TestCase):
    """
    Unit tests for the Event class.

    """

    def setUp(self):
        """
        Set up mock graph client for testing.

        """
        self.mock_client = GraphClient(
            Config(client_key_id="test", client_access_key="config")
        )
        self.test_event = Event(
            id="id1",
            patient_id="abc",
            start_time=42.1,
            end_time=None,
            classification={
                "namespace": "patient",
                "category": "activity",
                "enum": "testing",
            },
            display_name="Hello World",
            payload={"hello": "world"},
            method="manual",
            # NOTE: these attributes are returned in the dict representation,
            # but are not set as class attributes
            tags=[{"name": "test", "display_name": "Test"}],
            created_at=123,
            updated_at=456,
        )

    def test_attributes(self):
        """
        Test attributes for an initialized Event.

        """
        self.assertEqual(self.test_event.id, "id1")
        self.assertEqual(self.test_event.patient_id, "abc")
        self.assertEqual(self.test_event.display_name, "Hello World")
        self.assertEqual(self.test_event.start_time, 42.1)
        self.assertEqual(self.test_event.end_time, None)
        self.assertEqual(self.test_event.payload, {"hello": "world"})
        self.assertEqual(
            self.test_event.classification,
            {
                "namespace": "patient",
                "category": "activity",
                "enum": "testing",
            },
        )
        self.assertEqual(self.test_event.method, "manual")

    def test_to_dict(self):
        """Test dictionary representation"""
        self.assertEqual(
            self.test_event.to_dict(),
            {
                "display_name": "Hello World",
                "end_time": None,
                "id": "id1",
                "patient_id": "abc",
                "payload": {"hello": "world"},
                "start_time": 42.1,
                "classification": {
                    "namespace": "patient",
                    "category": "activity",
                    "enum": "testing",
                },
                "tags": [
                    {
                        "name": "test",
                        "display_name": "Test",
                    }
                ],
                "method": "manual",
                "created_at": 123,
                "updated_at": 456,
            },
        )

    def test_repr(self):
        """
        Test __repr__

        """
        self.assertEqual(
            'Event(name="Hello World", start_time=42.1)', repr(self.test_event)
        )

    def test__reformat_event(self):
        """
        Test _reformat_event

        """
        with self.subTest("no-end-time-max"):
            event = {
                "id": "id1",
                "patient_id": "abc",
                "display_name": "Sleep",
                "custom_detail": None,
                "payload": '{"hello": "world"}',
                "duration": {
                    "start_time": 42.1,
                    "end_time": 99,
                    "end_time_max": None,
                },
                "tags": [],
                "method": "manual",
            }
            _reformat_event(event)

            self.assertEqual(
                event,
                {
                    "id": "id1",
                    "patient_id": "abc",
                    "display_name": "Sleep",
                    "start_time": 42.1,
                    "end_time": 99,
                    "payload": {"hello": "world"},
                    "tags": [],
                    "method": "manual",
                },
            )

        with self.subTest("end-time-max"):
            event = {
                "id": "id1",
                "patient_id": "abc",
                "display_name": "Sleep",
                "custom_detail": {"display_name": ""},
                "payload": '{"hello": "world"}',
                "duration": {
                    "start_time": 42.1,
                    "end_time": 99,
                    "end_time_max": 100,
                },
                "method": "manual",
                "tags": [],
            }
            _reformat_event(event)

            self.assertEqual(
                event,
                {
                    "id": "id1",
                    "patient_id": "abc",
                    "display_name": "Sleep",
                    "start_time": 42.1,
                    "end_time": 100,
                    "payload": {"hello": "world"},
                    "method": "manual",
                    "tags": [],
                },
            )

        with self.subTest("custom-name"):
            event = {
                "id": "id1",
                "patient_id": "abc",
                "display_name": "Custom Event",
                "custom_detail": {"display_name": "Spoonful of Sugar"},
                "payload": '{"hello": "world"}',
                "duration": {
                    "start_time": 42.1,
                    "end_time": None,
                    "end_time_max": None,
                },
                "method": "manual",
                "tags": [],
            }
            _reformat_event(event)

            self.assertEqual(
                event,
                {
                    "id": "id1",
                    "patient_id": "abc",
                    "display_name": "Spoonful of Sugar",
                    "start_time": 42.1,
                    "end_time": None,
                    "payload": {"hello": "world"},
                    "method": "manual",
                    "tags": [],
                },
            )

        with self.subTest("tags"):
            event = {
                "id": "id1",
                "patient_id": "abc",
                "display_name": "Sleep",
                "custom_detail": None,
                "payload": '{"hello": "world"}',
                "duration": {
                    "start_time": 42.1,
                    "end_time": 99,
                    "end_time_max": None,
                },
                "tags": [{"name": "activity.aerobic"}, {"name": "activity.strength"}],
                "method": "manual",
            }
            _reformat_event(event)

            self.assertEqual(
                event,
                {
                    "id": "id1",
                    "patient_id": "abc",
                    "display_name": "Sleep",
                    "start_time": 42.1,
                    "end_time": 99,
                    "payload": {"hello": "world"},
                    "tags": ["activity.aerobic", "activity.strength"],
                    "method": "manual",
                },
            )

    def test_get_patient_events(self):
        """
        Test get_patient_events

        """
        self.mock_client.execute = mock.Mock(
            side_effect=[
                # empty response for the first "chunk" of time in the query range
                {"patient": {"eventList": {"events": []}}},
                {
                    "patient": {
                        "eventList": {
                            "events": [
                                {
                                    "id": "event1",
                                    "display_name": "Boxing",
                                    "custom_detail": None,
                                    "duration": {
                                        "start_time": 1671267538,
                                        "end_time": 1671268000,
                                        "end_time_max": None,
                                    },
                                    "payload": "",
                                    "classification": {
                                        "namespace": "patient",
                                        "category": "activity",
                                        "enum": "boxing",
                                    },
                                    "tags": [
                                        {
                                            "name": "activity.aerobic",
                                        },
                                        {
                                            "name": "activity.strength",
                                        },
                                    ],
                                    "method": "manual",
                                    "created_at": 1671268000,
                                    "updated_at": 1671268000,
                                }
                            ],
                            "pageInfo": {"endCursor": "cursor!"},
                        }
                    }
                },
                {
                    "patient": {
                        "eventList": {
                            "events": [
                                {
                                    "id": "event2",
                                    "display_name": "Custom Event",
                                    "custom_detail": {
                                        "display_name": "Spoonful of Sugar"
                                    },
                                    "duration": {
                                        "start_time": 1731267538.123,
                                        "end_time": None,
                                        "end_time_max": None,
                                    },
                                    "payload": '{"dosage": 1.0, "failed_dose": false}',
                                    "classification": {
                                        "namespace": "patient",
                                        "category": "medication",
                                        "enum": "custom-123",
                                    },
                                    "tags": [],
                                    "method": "manual",
                                    "created_at": 1731267538.123,
                                    "updated_at": 1731267538.456,
                                }
                            ]
                        }
                    }
                },
            ]
        )

        # Use a time range that is longer than the max.
        # The function should handle splitting it up into chunks.
        events = get_patient_events(
            patient_id="abc",
            start_time=1,
            end_time=10 + 90 * 24 * 60 * 60,
            client=self.mock_client,
        )

        expected = [
            {
                "patient_id": "abc",
                "display_name": "Boxing",
                "start_time": 1671267538,
                "end_time": 1671268000,
                "payload": {},
                "classification": {
                    "namespace": "patient",
                    "category": "activity",
                    "enum": "boxing",
                },
                "tags": ["activity.aerobic", "activity.strength"],
                "method": "manual",
                "created_at": 1671268000,
                "updated_at": 1671268000,
                "id": "event1",
            },
            {
                "patient_id": "abc",
                "display_name": "Spoonful of Sugar",
                "start_time": 1731267538.123,
                "end_time": None,
                "payload": {"dosage": 1.0, "failed_dose": False},
                "classification": {
                    "namespace": "patient",
                    "category": "medication",
                    "enum": "custom-123",
                },
                "tags": [],
                "method": "manual",
                "created_at": 1731267538.123,
                "updated_at": 1731267538.456,
                "id": "event2",
            },
        ]

        self.assertEqual(expected, events.to_list())

    def test_to_dataframe(self):
        """
        Test converting events to a dataframe

        """
        expected_columns = [
            "patient_id",
            "start_time",
            "end_time",
            "namespace",
            "category",
            "enum",
            "display_name",
            "method",
            "payload",
            "tags",
            "created_at",
            "updated_at",
            "id",
        ]

        # Make sure empty dataframe works
        empty_df = EventSet([]).to_dataframe()
        self.assertListEqual(list(empty_df.columns), expected_columns)

        # Test with a few events
        ev2 = Event(
            id="id2",
            patient_id="abc",
            start_time=100,
            end_time=140,
            classification={
                "namespace": "patient",
                "category": "activity",
                "enum": "healthkit-running",
            },
            display_name="Hello World",
            payload={"hello": "world"},
            method="healthkit",
            tags=[],
            created_at=123,
            updated_at=456,
        )

        # same start time as ev2, but an earlier end time
        ev3 = Event(
            id="id3",
            patient_id="abc",
            start_time=100,
            end_time=100,
            classification={
                "namespace": "patient",
                "category": "note",
            },
            display_name="Hello World",
            payload={"hello": "world"},
            method="manual",
            tags=[],
            created_at=123,
            updated_at=456,
        )

        # add events to the event set out of order (they should be sorted
        # by time in the dataframe)
        event_set = EventSet([ev3, self.test_event, ev2])

        df = event_set.to_dataframe()
        self.assertListEqual(list(df.columns), expected_columns)

        # check ids to make sure sorting was done correctly
        self.assertEqual(list(df.id), ["id1", "id3", "id2"])

        # check that classification was expanded into columns
        self.assertEqual(list(df.namespace), ["patient", "patient", "patient"])
        self.assertEqual(list(df.category), ["activity", "note", "activity"])
        self.assertEqual(list(df.enum), ["testing", None, "healthkit-running"])

    def test_event_type_helpers(self):
        """Minimal test to exercise the helpers that fetch one type of event.

        Only asserts that the expected filters were passed to the GraphQL client.

        """
        self.mock_client.execute = mock.Mock(
            return_value={"patient": {"eventList": {"events": []}}},
        )

        # activity
        get_patient_activity_events(
            "abc", start_time=1, end_time=10, client=self.mock_client
        )
        self.mock_client.execute.assert_called_once_with(
            statement=mock.ANY,
            # GraphQL variables
            patient_id="abc",
            cursor=None,
            start_time=1,
            end_time=10,
            include_filters=[
                {
                    "namespace": "patient",
                    "category": "activity",
                    "enum": "*",
                }
            ],
        )

        # medication
        self.mock_client.execute.reset_mock()

        get_patient_medication_events(
            "abc", start_time=1, end_time=10, client=self.mock_client
        )

        self.mock_client.execute.assert_called_once_with(
            statement=mock.ANY,
            # GraphQL variables
            patient_id="abc",
            cursor=None,
            start_time=1,
            end_time=10,
            include_filters=[
                {
                    "namespace": "patient",
                    "category": "medication",
                    "enum": "*",
                }
            ],
        )

        # symptom
        self.mock_client.execute.reset_mock()

        get_patient_symptom_events(
            "abc", start_time=1, end_time=10, client=self.mock_client
        )

        self.mock_client.execute.assert_called_once_with(
            statement=mock.ANY,
            # GraphQL variables
            patient_id="abc",
            cursor=None,
            start_time=1,
            end_time=10,
            include_filters=[
                {
                    "namespace": "patient",
                    "category": "symptom",
                    "enum": "*",
                }
            ],
        )

        # wellbeing
        self.mock_client.execute.reset_mock()

        get_patient_wellbeing_events(
            "abc", start_time=1, end_time=10, client=self.mock_client
        )

        self.mock_client.execute.assert_called_once_with(
            statement=mock.ANY,
            # GraphQL variables
            patient_id="abc",
            cursor=None,
            start_time=1,
            end_time=10,
            include_filters=[
                {
                    "namespace": "patient",
                    "category": "wellbeing",
                    "enum": "*",
                }
            ],
        )



"""
Events are created by users of the StrivePD app. They fall into
many different categories, including:

    - Activities (manually logged or ingested from HealthKit)
    - Medication and supplement logs
    - Symptom logs
    - Wellbeing logs
    - Free text notes
    - etc.

Utility functions are provided to query events of specific types.
Other event categories may exist: use the get_patient_events function to
fetch events of any type.

NOTE: Many StrivePD events are also queryable as **streams** (with the algorithm `ingest-rune-events`).
The stream representation of the data is **less reliable** and may not reflect the latest state
of user data. **We recommend using this module to query StrivePD Events, whenever possible.**

"""

import json
from typing import Iterable, Optional, Type

import pandas as pd

from .client import GraphClient, global_graph_client
from .common import ItemBase, ItemSet
from .internal import _time_type, _time_type_to_unix_secs


class Event(ItemBase):
    """
    A single event.

    """

    def __init__(
        self,
        id: str,
        patient_id: str,
        start_time: float,
        end_time: Optional[float],
        classification: dict,
        display_name: str,
        payload: Optional[dict] = None,
        method: Optional[str] = None,
        **attributes,
    ):
        """
        Initialize an Event: an item recorded by a user of the StrivePD app.

        Args:
            id: Event ID
            patient_id: Patient ID
            start_time: Start time of the event
            end_time: End time of the event
            classification: Event classification (see below for details).
            display_name: display name of the event
            payload: Event payload
            **attributes: additional event metadata (included in the dict
                representation of the Event).

        Event classification is a three-level hierarchy, represented as a
        dictionary with 2 or 3 keys:

            - **namespace**: the general source of the event (e.g. "patient")
            - **category**: the type of event (e.g. "activity", "medication")
            - **enum**: a specific enumeration within the category (e.g. "running").
              Not all events have an enumeration.

        Note that this classification uses similar (but not identical!) terminology to
        the events that are queryable through the V2 Stream API.

        """
        self.patient_id = patient_id
        self.start_time = start_time
        self.end_time = end_time

        # validate classification
        expected_keys = {"namespace", "category"}
        missing_keys = expected_keys - classification.keys()
        if missing_keys:
            raise ValueError(f"Classification missing required keys: {missing_keys}")

        self.classification = classification

        self.display_name = display_name
        self.payload = payload or {}
        self.method = method

        super().__init__(
            id=id,
            patient_id=patient_id,
            start_time=start_time,
            end_time=end_time,
            classification=classification,
            display_name=display_name,
            payload=payload,
            method=method,
            **attributes,
        )

    def __repr__(self):
        """
        Override repr, to include start time and name

        """
        return (
            f"{self.__class__.__name__}("
            f'name="{self.display_name}", start_time={self.start_time})'
        )


class EventSet(ItemSet):
    """
    A collection of Events.

    """

    def __init__(self, items: Iterable[Event] = ()):
        """
        Initialize with Events.

        """
        super().__init__(items=items)

    @property
    def _item_class(self) -> Type[ItemBase]:
        """
        Instance type of items in this set.

        """
        return Event

    def to_dataframe(self) -> pd.DataFrame:
        """
        Return a dataframe of events, sorted by time.

        """
        df = super().to_dataframe()

        if not df.empty:
            df = df.sort_values(by=["start_time", "end_time"])
            df = df.reset_index(drop=True)

            # parse classification into fields (for easier querying)
            df["namespace"] = df.classification.apply(lambda x: x.get("namespace"))
            df["category"] = df.classification.apply(lambda x: x.get("category"))
            df["enum"] = df.classification.apply(lambda x: x.get("enum"))

            # parse datetime columns
            df["start_time"] = pd.to_datetime(df["start_time"], unit="s")
            df["end_time"] = pd.to_datetime(df["end_time"], unit="s")
            df["created_at"] = pd.to_datetime(df["created_at"], unit="s")
            df["updated_at"] = pd.to_datetime(df["updated_at"], unit="s")

        # Reorder columns (dropping classification)
        df = df.reindex(
            columns=[
                "patient_id",
                "start_time",
                "end_time",
                "namespace",
                "category",
                "enum",
                "display_name",
                "method",
                "payload",
                "tags",
                "created_at",
                "updated_at",
                "id",
            ]
        )

        return df


# GraphQL query to get patient events
_EVENT_GQL_QUERY = """
query getEventList(
    $patient_id: ID!,
    $cursor: Cursor,
    $start_time: Float!,
    $end_time: Float!,
    $include_filters: [EventClassificationFilter],
) {
    patient(id: $patient_id) {
        eventList(
            startTime: $start_time,
            endTime: $end_time,
            cursor: $cursor,
            includeFilters: $include_filters,
        ) {
            events {
                id
                display_name: displayName
                custom_detail: customDetail {
                    display_name: displayName
                }
                duration {
                    start_time: startTime
                    end_time: endTime
                    end_time_max: endTimeMax
                }
                payload
                classification {
                    namespace
                    category
                    enum
                }
                tags {
                    name
                }
                method
                created_at: createdAt
                updated_at: updatedAt
            }
            pageInfo {
                endCursor
            }
        }
    }
}
"""

# Each event query may only fetch up to 90 days of data
MAX_QUERY_RANGE_SECS = 90 * 24 * 60 * 60


def _iter_events(
    patient_id: str,
    start_time: _time_type,
    end_time: _time_type,
    include_filters: Optional[list] = None,
    client: Optional[GraphClient] = None,
):
    """Query events for a patient and iterate over the results.
    Formats each event as kwargs that can be passed to Event.

    """
    graph_client = client or global_graph_client()
    start_time = _time_type_to_unix_secs(start_time)
    end_time = _time_type_to_unix_secs(end_time)

    next_cursor = None

    variables = {}
    if include_filters:
        variables["include_filters"] = include_filters

    # query 90 days of events at a time
    current_start_time = start_time
    while current_start_time < end_time:
        current_end_time = min(current_start_time + MAX_QUERY_RANGE_SECS, end_time)

        result = graph_client.execute(
            statement=_EVENT_GQL_QUERY,
            # GraphQL variables
            patient_id=patient_id,
            cursor=next_cursor,
            start_time=current_start_time,
            end_time=current_end_time,
            **variables,
        )

        event_list = result["patient"]["eventList"]
        for event in event_list.get("events", []):
            _reformat_event(event)
            yield Event(patient_id=patient_id, **event)

        # when pagination is exhausted for this time range, move to the
        # next time window (if any)
        next_cursor = event_list.get("pageInfo", {}).get("endCursor")
        if not next_cursor:
            current_start_time = current_end_time


def _reformat_event(event: dict):
    """Reformat event in place, so it can be passed as kwargs to initialize Event."""
    # replace duration with start and end time
    duration = event.pop("duration") or {}
    event["start_time"] = duration.get("start_time")

    end_time = duration["end_time"]
    end_time_max = duration["end_time_max"]
    if end_time_max is not None:
        end_time = end_time_max

    event["end_time"] = end_time

    # set a single display name, using custom_detail if available
    custom_detail = event.pop("custom_detail") or {}
    custom_name = custom_detail.get("display_name")
    if custom_name:
        event["display_name"] = custom_name

    # parse the payload as JSON
    payload = event.pop("payload") or "{}"
    event["payload"] = json.loads(payload)

    # flatten tag names into a list
    tags = event.pop("tags") or []
    event["tags"] = [tag["name"] for tag in tags]


def get_patient_events(
    patient_id: str,
    start_time: _time_type,
    end_time: _time_type,
    include_filters: Optional[list] = None,
    client: Optional[GraphClient] = None,
) -> EventSet:
    """Fetch a patient's Events, in an optional time range.

    Args:
        patient_id: Patient ID
        start_time: Optional start time for the query range
        end_time: Optional end time for the query range
        include_filters: Event classification filters to include. If
            this is not specified, all events are returned.
        client: GraphClient instance (optional)

    """
    events = _iter_events(
        patient_id=patient_id,
        start_time=start_time,
        end_time=end_time,
        include_filters=include_filters,
        client=client,
    )

    return EventSet(events)


def get_patient_activity_events(
    patient_id: str,
    start_time: _time_type,
    end_time: _time_type,
    client: Optional[GraphClient] = None,
) -> EventSet:
    """Fetch a patient's activity Events.

    This includes activities that were manually logged through the
    StrivePD app and activities ingested from HealthKit (which can
    come from third party apps or devices).

    """
    include_filters = [
        {
            "namespace": "patient",
            "category": "activity",
            "enum": "*",
        }
    ]
    return get_patient_events(
        patient_id=patient_id,
        start_time=start_time,
        end_time=end_time,
        include_filters=include_filters,
        client=client,
    )


def get_patient_medication_events(
    patient_id: str,
    start_time: _time_type,
    end_time: _time_type,
    client: Optional[GraphClient] = None,
) -> EventSet:
    """Fetch a patient's medication Events, as recorded through
    the StrivePD app.

    NOTES:

    - There are several ways to log a medication in StrivePD,
      including an "autolog" feature that automatically records a
      medication on a schedule. The "method" attribute of the Event
      object indicates how the medication event was created.
    - The display name for custom medications is only queryable by
      certain user roles. These event logs will default to a generic
      display name.

    """
    include_filters = [
        {
            "namespace": "patient",
            "category": "medication",
            "enum": "*",
        },
    ]
    return get_patient_events(
        patient_id=patient_id,
        start_time=start_time,
        end_time=end_time,
        include_filters=include_filters,
        client=client,
    )


def get_patient_symptom_events(
    patient_id: str,
    start_time: _time_type,
    end_time: _time_type,
    client: Optional[GraphClient] = None,
) -> EventSet:
    """Fetch a patient's symptom Events, as recorded through
    the StrivePD app.

    """
    include_filters = [
        {
            "namespace": "patient",
            "category": "symptom",
            "enum": "*",
        },
    ]
    return get_patient_events(
        patient_id=patient_id,
        start_time=start_time,
        end_time=end_time,
        include_filters=include_filters,
        client=client,
    )


def get_patient_wellbeing_events(
    patient_id: str,
    start_time: _time_type,
    end_time: _time_type,
    client: Optional[GraphClient] = None,
) -> EventSet:
    """Fetch a patient's wellbeing Events, as recorded through
    the StrivePD app.

    """
    include_filters = [
        {
            "namespace": "patient",
            "category": "wellbeing",
            "enum": "*",
        }
    ]
    return get_patient_events(
        patient_id=patient_id,
        start_time=start_time,
        end_time=end_time,
        include_filters=include_filters,
        client=client,
    )
